/* 
 * Copyright (c) 1990, 1991, 1994 The University of Utah and
 * the Computer Systems Laboratory at the University of Utah (CSL).
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: cio.c 1.3 94/11/10$
 *	Author: Bob Wheeler, Leigh Stoller, University of Utah CSL
 */

/*
 *	CIO driver for hp800 series machines
 */

#include <cio.h>
#if NCIO > 0

#include <sys/types.h>
#include <device/param.h>
#include <device/errno.h>
#include <device/buf.h>
#include <device/conf.h>
#include <machine/machparam.h>
#include <machine/spl.h>

#include <machine/iodc.h>
#include <machine/iomod.h>
#include <machine/pdc.h>

#include <hpcio/device.h>
#include <hpcio/cioreg.h>

static int prepchain();
caddr_t cio_malloc();

extern struct iomod *prochpa;

/*
 * function to convert kernel addresses into physical addresses and
 * flush kernel addresses from the data cache
 */
#define	phys_addr(address) (ltor(0,(caddr_t) (address)))
#define flush_data_cache(address, count) fdcache(0, (address), (count))
#define purge_data_cache(address, count) pdcache(0, (address), (count))

void	ciointr();

int	cioattach(), cioprobe(), cioscan(), ciolayer();
void	ciodisplay();
struct	driver ciodriver = {
	"cio", cioattach, cioprobe, cioscan, ciodisplay, ciolayer
};

struct cio_softc {
	struct modtab	   *modtab;	/* modtab structure for this device */
	struct ciohpa	   *hpa;	/* hard physical address for CIO */
	struct ciospa	   *spa;	/* soft physical address for CIO */
	int		   eim;		/* external interrupt installed at */
	int		   max_channel;	/* maximum channel used on CIO */
	int		   alive;	/* autoconfigured? */
	struct cio_request *cr;		/* CIO request structure */
	struct dmaquad     *cmdq;	/* general command quad */
	struct linkquad    *link;	/* link quad for completion */
	struct semaquad    *sema;	/* semaphore quad */
	struct statquad    *stat;	/* status quad */
	struct cio_chan    cio_chan[MAX_CIO_CHAN]; 
};

struct cio_softc cio_softc[NCIO];
struct cio_request ciocr[NCIO];

/*
 * cio probe routine
 */
cioprobe(m)
	struct modtab *m;
{
	return (m->m_type.iodc_type == IODC_TP_CIO) ? 0 : ENXIO;
}

/* 
 * attach and initialize CIO bus
 */
cioattach(m, unit)
	struct modtab *m;
{
	register struct cio_softc *cs = &cio_softc[unit];
	caddr_t chain;
	struct dmaquad *q;
	int chan;

	if (unit >= NCIO)
		return ENXIO;

	if (cs->alive)
		return 0;

	cs->modtab = m;
	cs->hpa    = (struct ciohpa *) m->m_hpa;
	cs->spa    = (struct ciospa *) m->m_spa;
	cs->eim    = SPL_IBIT_CIO;
	cs->max_channel = 0;
	cs->cr     = &ciocr[unit];

	/*
	 * Allocate chain ram for the device and flush to memory.
	 */
	chain = (caddr_t) cio_malloc(CIO_CHAIN_SIZE, CIO_CHAIN_ALIGN);
	bzero(chain, CIO_CHAIN_SIZE);
	flush_data_cache(chain, CIO_CHAIN_SIZE);

	(void)procitab(SPLCIO, ciointr, 0, 0);

	/* 
	 * set eim and chain addr
	 */
	io_command(m->m_hpa, CMD_RESET);
	cs->hpa->chain_ram_base = (caddr_t) chain;
	/*
	 * XXX - This next line is assuming that all CIOs interrupt through
	 * the monarch processor. Needs to be changed for MP.
	 */
	cs->hpa->io_eim = cs->eim | (u_int) PAGE0->mem_hpa;
	cs->hpa->io_spa_address = IOSPA(m->m_spa, m->m_type);

	/*
	 * clear the cards internal memory
	 */
	for (chan = 0; chan < MAX_CIO_CHAN; chan++) {
		cs->spa->subc[chan].status =  0x00100000;
		cs->spa->subc[chan].control = 0x10000000;
	}

	/*
	 * Set up the HPA registers
	 */
	cs->hpa->sub_mask_clr = 0xffffffff;
	cs->hpa->interrupt    = 0xffffffff;
	cs->hpa->nmi_address  = (caddr_t) &PAGE0->mem_hpa->io_command;
	cs->hpa->nmi_data     = CMD_RESET;

	/*
	 * wait 4 seconds for the CIO before issuing commands
	 */
	delay(4000000);

	/*
	 * assign addresses to the channels
	 */
	for (chan = 0; chan < MAX_CIO_CHAN; chan++)
		cs->spa->direct[chan].data.normal = 1 << chan;

	/*
	 * Assign address spaces and do general initialization of channels
	 */
	for (chan = 0; chan < MAX_CIO_CHAN; chan++) {
		register struct cio_chan *cc = &cs->cio_chan[chan];
		cc->type    = 0;
		cc->ciounit = unit;
		cc->channel = chan;
		cc->alive   = 0;
		cc->device  = (struct bus *)0;
		cc->dio     = &cs->spa->direct[chan];
		cc->subc    = &cs->spa->subc[chan];
		cc->chain   = chain + CIO_QUAD_SIZE * chan;
		cc->hpa     = cs->hpa;
		cc->active  = (struct cio_request *)0;
		cc->debug   = 0;
	}

	/*
	 * initialize quads to do identify command on channels
	 */
       	q = (struct dmaquad *) cio_malloc(sizeof(struct dmaquad) * 4, 
					  CIO_BUFFER_ALIGN);
	cs->cmdq = q;
	cs->link = (struct linkquad *) (q + 1);
	cs->sema = (struct semaquad *) (q + 2);
	cs->stat = (struct statquad *) (q + 3);

	cs->cmdq->link    = (struct dmaquad *) cs->link;
	cs->cmdq->command = CIODMA_IDY;
	cs->cmdq->address = (caddr_t) cio_malloc(sizeof(struct idy), 
						 CIO_BUFFER_ALIGN);
	cs->cmdq->count   = sizeof(struct idy);

	cs->link->link    = CIO_EOC(cs->link);
	cs->link->command = CIODMA_LNKST;
	cs->link->sema    = cs->sema;
	cs->link->stat    = cs->stat;

	/*
	 * initialize CIO request structure
	 */
	cs->cr->cmdq      = cs->cmdq;
	cs->cr->stat      = cs->stat;
	cs->cr->quad      = (struct dmaquad *)0;
	cs->cr->ciounit   = unit;

	cs->alive = 1;

	return 0;
}

/*
 * scan a CIO bus for attached devices
 */
cioscan(unit)
{
	struct cio_softc *cs = &cio_softc[unit];
	int chan;
	extern struct bus bus_devices[];
	struct bus *device;


	if (unit >= NCIO || !cs->alive)
		return ENXIO;

	/*
	 * scan the CIO and see what channels have something in them
	 */
	for (chan = 0; chan < MAX_CIO_CHAN; chan++) {
		int status;
		register struct cio_chan *cc = &cs->cio_chan[chan];
		register struct idy *idy = (struct idy *) cs->cmdq->address;

		status = cc->dio->ctlsens.normal & 0xff;

		/*
		 * if the present bit isn't set then the device either isn't 
		 * there or it's a level 1 device and we don't know what to do
		 * with it.
		 */
		if (!(status & CIO_SENSE_PRE))
			continue;

		/*
		 * This channel has something in it that is not a level 1 
		 * device. Connect a subchannel for this slot then enable and
		 * initialize it so that we can interrogate the device and 
		 * find out what it is.
		 */
		cc->dio->statcmd.normal = CIO_SUBC_CMD_CSC | chan;
		cc->dio->ctlsens.normal = CIO_DA_CTRL_ARE;

		/*
		 * start the DMA chain for the channel and wait until done
		 */
		cs->cr->channel = chan;
		if (ciosync(cs->cr)) {
			printf("cioscan: can't determine identity of card in cio%d slot %d\n", unit, chan);
			continue;
		}

		/*
		 * save the type of device on this channel
		 */
		cc->type = idy->card_id;
	}

	/*
	 * look in the configuration table and try to identify the controllers
	 */
	for (device = bus_devices; device->driver; device++) {
		/*
		 * if this device isn't on a CIO then skip it
		 */
		if (device->bustype != &ciodriver)
			continue;

		/*
		 * see if this is the right unit or if it's a wildcard CIO unit
		 */
		if (device->busunit != -1 && device->busunit != unit)
			continue;

		/*
		 * This is the right CIO unit search all channels for something
		 * that looks like what we want...
		 */
		for (chan = 0; chan < MAX_CIO_CHAN; chan++) {
			struct cio_chan *cc = &cs->cio_chan[chan];

			/*
			 * see if this controller has already been configured 
			 * doesn't exist or isn't in the right slot
			 */
			if (cc->alive || !cc->type || 
			    (device->busslot != chan && device->busslot != -1))
				continue;

			/*
			 * see if we recognize this device as what we want...
			 */
			if ((*device->driver->probe)(cc) != 0)
				continue;
				
			/*
			 * this looks like it... attach it and associate it 
			 * with the right unit
			 */
			if ((*device->driver->attach)
			    (cc, device->unit,device->flags) != 0)
				continue;

			/*
			 * Now scan it for devices or controllers below it
			 */
			if ((*device->driver->scan)(device->unit) != 0)
				continue;

			/*
			 * found it and initialized it
			 */
			if (chan > cs->max_channel)
				cs->max_channel = chan;

			cc->alive = 1;
			cc->device = device;
			break;
		}
	}
	return 0;
}

void
ciodisplay(unit)
{
	struct cio_softc *cs = &cio_softc[unit];
	int chan;
	extern int indent;

	for (chan = 0; chan < MAX_CIO_CHAN; chan++) {
		struct cio_chan *cc = &cs->cio_chan[chan];
		if (!cc->type)
			continue;

		if (cc->alive) {
			iprintf("%s%d in cio%d slot %d\n", 
				cc->device->driver->name,
				cc->device->unit, unit, chan);	
			indent += 5;
			(*cc->device->driver->display)(cc->device->unit);
			indent -= 5;
		} else
			iprintf("unknown device id %d in cio%d slot %d\n",
				cc->type, unit, chan);
	}
}

ciolayer(unit, layer, devname)
	int *layer;
	char *devname;
{
	struct cio_softc *cs = &cio_softc[unit];
	struct cio_chan *cc = &cs->cio_chan[*layer];

	if (*layer >= MAX_CIO_CHAN)
		return ENXIO;

	if (!cc->type)
		return ENXIO;

	if (cc->alive)
		return (*cc->device->driver->layer)
			(cc->device->unit, layer + 1, devname);

	return ENXIO;
}

check_cio_status(cc)
	struct cio_chan *cc;
{
	int status;
	int mstat;

	status = cc->hpa->io_status;

	if (status & IO_STATUS_ME) {
		printf("cio%d module error - ", cc->ciounit);

		switch(mstat = status & IO_STATUS_MSTAT) {
		case MERR_DEPEND:
			printf("internal catastrophic channel error\n");
			break;

		case MERR_LOOPBACK:
			printf("loopback error\n");
			break;

		case MERR_DATA:
			printf("data parity error\n");
			break;

		case MERR_PROTOCOL:
			printf("bus protocol error\n");
			break;

		case MERR_ADDRESS:
			printf("slave did not respond to address\n");
			break;

		case MERR_NOARQ:
			printf("ARQ asserted without response to poll\n");
			break;
		default:
			printf("unknown error, mstat = 0x%x\n", mstat);
			break;
		}

		if (mstat != MERR_DATA) {
			printf("reseting cio%d\n", cc->ciounit);
			cioreset(cc->ciounit);
			return EIO;
		}
	}

	return 0;
}


/*
 * do a direct cio command and check for errors. We assume that the CIO and 
 * channel are alive and well.
 */
ciocommand(cc, command)
	struct cio_chan *cc;
{
	int status;
	int mstat;
	int count;

	/*
	 * If the ready bit isn't on it will be shortly, in any case check 
	 * for a device timeout. This register is in HPA space (I/O space)
	 * and therefor it's not cached.
	 */
	count = 0;
	while(!(cc->hpa->io_status & IO_STATUS_RY) && count < CIO_TIME_OUT) {
		delay(10);
		count++;
	}

	if (count == CIO_TIME_OUT) {
		printf("ciocommand: timed out waiting for CIO ready bit\n");
		return EIO;
	}
		
	cc->hpa->io_command = cc->channel << 8 | command;
	cc->hpa->sub_mask_set = 1 << cc->channel;

	/*
	 * the hardware will ensure that we wait until the command is
	 * completed before it returns the status
	 */
	return (check_cio_status(cc));
}

/*
 * This routine will issue a cio request and then spin wait for it to complete.
 * This should only be used when we know that data is present on a device. It 
 * is equivalent to ciogo() but synchronous. Calls to this routine should not
 * set the the interrupt bit in any command and need not set the done or 
 * unit fields in the cio_request structure.
 */
ciosync(cr)
	struct cio_request *cr;
{
	struct cio_chan *cc = &cio_softc[cr->ciounit].cio_chan[cr->channel];
	int status;
	int count;

	if (cr->ciounit > NCIO || cr->channel > MAX_CIO_CHAN) {
		printf("ciosync: bad unit number\n");
		return ENXIO;
	}

	if (cc->debug) {
		printf("\n\n------ ciosync dumping request -------\n");
		dump_cr(cr);
		dump_cc(cc);
		dump_quads(cr->cmdq);
		dump_status(cc);
	}

	if ((status = prepchain(cr)) != 0)
		return status;

	/*
	 * wait for this channel to accept a command, check for timeouts
	 */
	count = 0;
	while (!(cc->subc->status & SUBC_STAT_RDY) && count < CIO_TIME_OUT) {
		status = cc->subc->status;
		delay(10);
		count++;
	}

	if (count == CIO_TIME_OUT) {
		printf("ciosync: timed out waiting for subchannel ready\n");
		return EIO;
	}

	/*
	 * copy the request quad to chain ram and flush the cache
	 */
	bcopy(cr->cmdq, cc->chain, CIO_QUAD_SIZE);
	flush_data_cache(cc->chain, CIO_QUAD_SIZE);

	/*
	 * start the chain on the CIO
	 */
	if ((status = ciocommand(cc, CMD_CHAIN))) {
		printf("ciosync: failed ciocommand\n");
		return status;
	}

	/*
	 * wait for the DMA to complete and the status flag to be set.
	 */
	count = 0;
	purge_data_cache(cr->stat, CIO_QUAD_SIZE);
	while (!cr->stat->link && count < CIO_TIME_OUT) {
		delay(10);
		purge_data_cache(cr->stat, CIO_QUAD_SIZE);
		count++;
	}

	if (count == CIO_TIME_OUT) {
		printf("ciosync: timed out waiting for completion\n");
		dump_status(cc);
		return EIO;
	}

	/*
	 * since this command was syncronous we can check if any errors
	 * occured on the card.
	 */
	if ((status = check_cio_status(cc)) != 0)
		return status;

	if (cr->polled) {
		printf("setting sub_mask_set...\n");
		cc->hpa->sub_mask_set = 1 << cr->channel;
		dump_status(cc);
	}

	return 0;
}


/*
 * Start an DMA chain on the CIO. This device starts the cio request and
 * returns asynchronously. When the command has finished the done routine
 * specified in the request structure will be called. The quad chain should
 * terminate with a link quad that has the interrupt bit set in the command.
 * The semaphore quad should be cleared, the semaphore quad sema_head field
 * should be zero and the entire chain flushed to memory.
 * 
 */
ciogo(cr)
	struct cio_request *cr;
{
	struct cio_chan *cc = &cio_softc[cr->ciounit].cio_chan[cr->channel];
	int status;
	int count;

	if (cr->ciounit > NCIO || cr->channel > MAX_CIO_CHAN)
		return ENXIO;

	if (!cc->alive)
		return ENXIO;

	if (cc->debug) {
		printf("\n\n------ ciogo dumping request -------\n");
		dump_cr(cr);
		dump_cc(cc);
		dump_quads(cr->cmdq);
		dump_status(cc);
	}

	/*
	 * we only allow one request per channel to be active
	 */
	if (cc->active != (struct cio_request *)0)
		return EIO;

	if ((status = prepchain(cr)) != 0)
		return status;

	/*
	 * this is now the active request
	 */
	cc->active = cr;

	/*
	 * wait for this channel to accept a command, It should already
	 * be ready, just make sure...
	 */
	count = 0;
	while (!(cc->subc->status & SUBC_STAT_RDY) && count < CIO_TIME_OUT) {
		delay(10);
		count++;
	}

	if (count == CIO_TIME_OUT)
		return EIO;

	/*
	 * copy the request quad to chain ram and flush the cache
	 */
	bcopy(cr->cmdq, cc->chain, CIO_QUAD_SIZE);
	flush_data_cache(cc->chain, CIO_QUAD_SIZE);

	/*
	 * start the request, if there is an error then clear the active
	 * request field
	 */
	if ((status = ciocommand(cc, CMD_CHAIN)) != 0)
		cc->active = (struct cio_request *)0;

	return status;
}

/*
 * CIO interrupt service routine
 */
void
ciointr()
{
	int status;
	int chan;
	int unit;
	struct cio_chan    *cc;
	struct cio_request *cr;
	struct cio_softc   *cs;

	for (unit = 0; unit < NCIO; unit++) {
		cs = &cio_softc[unit];
		
		if (!cs->alive)
			continue;

		/*
		 * Scan all channels and see if any DMA chains are finished
		 */
		for (chan = 0; chan <= cs->max_channel; chan++) {
			cc = &cs->cio_chan[chan];

			if (!cc->alive ||cc->active == (struct cio_request *)0)
				continue;

			cr = cc->active;
			
			/*
			 * Every cio command must specify a status quad to 
			 * test for completion. If the link in the status 
			 * quad is non-zero then this is an operation that 
			 * has completed.
			 */
			if (!cr->stat->link) {
				purge_data_cache(cr->stat, CIO_QUAD_SIZE);
				continue;
			}

			/*
			 * this DMA request has completed, schedule an 
			 * interrupt at the appropriate priority
			 */
			set_interrupt(cc->int_level);
		}
	
		/*
		 * Now check the interrupt status register for errors and ARQs
		 */
		status = cs->hpa->interrupt;

		if (status) {
			for (chan = 0; chan <= cs->max_channel; chan++) {
				if ((status & 0x00010001 << chan) == 0)
					continue;

				cc = &cs->cio_chan[chan];

				if (!cc->alive) {
					printf("spurious interrupt on cio%d channel %x\n", 
					       unit, chan);
					cc->hpa->interrupt = 
						0x00010001 << chan;
					cc->dio->ctlsens.normal = 
						CIO_DA_CTRL_ARE; 
					continue;
				}

				if (cc->notified)
					continue;

				cc->notified = 1;
				set_interrupt(cc->int_level);

				if (cc->dio->ctlsens.normal & CIO_SENSE_ARQ)
					status = cc->dio->statcmd.normal;
				else
					printf("chan %d not at ARQ\n", chan);
			}
		}
	}
}

dequeue_cio_request(cc)
	struct cio_chan *cc;
{
	int s;

	s = splbio();
	cc->active = (struct cio_request *)0;
	splx(s);
}

test_cio_are(cc) 
	struct cio_chan *cc;
{
	int status;
	int s;

	s = splbio();
	status = cc->hpa->interrupt & (0x00010001 << cc->channel);
	splx(s);

	return status;
}

clear_cio_are(cc)
	struct cio_chan *cc;
{
	int s;
	s = splbio();

	cc->hpa->interrupt = 0x00010001 << cc->channel;
	cc->dio->ctlsens.normal = CIO_DA_CTRL_ARE; 
	cc->notified = 0;

	splx(s);
}


/*
 * XXX - should at least try....
 */
cioreset(ciounit)
{
	panic("reset cio%d\n", ciounit);
}


/*
 * prepchain
 *
 * This routine prepares quads for DMA I/O. It makes two passes over the chain.
 * In the first pass it checks for data buffers that cross page boundaries and
 * operations that require byte mode transfers. If a buffer crosses a page 
 * boundary and the fg pointer is not NULL, additional dmaquads are inserted 
 * into the chain to perform the scatter/gather operations on the physical 
 * pages. The fg pointer should point to an array of dmaquads sufficient to 
 * map all the scatter/gather requests. No check is made for running out of 
 * additional dmaquads pointed to by fg. If the fg pointer is 0 scatter/gather
 * checks are not made for the chain.
 *
 * In the second pass over the chain alignment checks are made for all pointers
 * and the semaphore in any semaquad is set to 1. The value of head_next in any
 * semaphore quad is set to zero. If any quads needed byte mode transfers then
 * all quads in the chain are set to use byte mode. All pointers in the chain 
 * are converted to physical addresses. Then flush all quads and data buffers 
 * from the cache.
 *
 * Each dmaquad has two identical set of pointers, commands and counts, one for
 * physical addresses and one ofr virtual addresses. When we insert scatter
 * gather quads we do not change the virtually addressed set of pointers. This
 * way the same virtually set up quad chain can be used over and over without 
 * having to get rid of specific scatter gather chains. When we convert from 
 * virtual to physical addresses we of course only change the physical 
 * addresses in the quad.
 *
 * The routine returns 0 on success and EINVAL for other errors.
 */
static int
prepchain(cr)
	struct cio_request *cr;
{
	struct cio_chan *cc = &cio_softc[cr->ciounit].cio_chan[cr->channel];
	struct dmaquad *q = cr->cmdq;
	struct dmaquad *fq = cr->quad;
	int 	bytemode = 0;
	caddr_t address;
	int 	count;
	int 	available;
	struct	dmaquad *p;
	struct 	dmaquad *next_p;
	int 	end_of_chain;
	struct dmaquad *next_link;

	/*
	 * First pass...
	 *
	 * Look for byte transfers, illegal counts and scatter/gather problems
	 */
	for (p = q; !((u_int) p & 1); p = next_link) {

		/*
		 * when we add scatter/gather quads we lose the link pointer,
		 * save it here.
		 */
		next_link = p->link;

		/*
		 * copy the virtual addresses into the physical addresses
		 * so that we can add scatter/gather buffers and then convert
		 * the addresses.
		 */
		p->plink    = p->link;
		p->pcommand = p->command;
		p->paddress = p->address;
		p->pcount   = p->count;

		/*
		 * we don't need to do anything to LNKST, PAU or DIS commands
		 */
		if (IS_CIODMA_LNKST(p->command) || IS_CIODMA_PAU(p->command) ||
		    IS_CIODMA_DIS(p->command))
			continue;

		if (p->pcount <= 0) {
			printf("prepchain: count invalid\n");
			return EINVAL;
		}

		/*
		 * check for byte mode transfer
		 */
		if (p->pcommand & CIO_BYTE || p->pcount & 1)
			bytemode = 1;

		/*
		 * if we don't have a list of additional quads then don't 
		 * check scatter gather requirements.
		 */
		if (fq == (struct dmaquad *)0)
			continue;

		/*
		 * Check for scatter/gather problems
		 */
		address = p->paddress;
		count   = p->pcount;

		while(1) {
			p->paddress = address;
			available = HP800_PGBYTES -
				((int) address & HP800_PGBYTES - 1);

			/*
			 * if we can fit the remaining count into this quad 
			 * then do it and get out of here...
			 */
			if (count <= available) {
				p->pcount = count;
				break;
			}

			/*
			 * can't fit it all, take all we can get
			 */
			p->pcount = available;
			count -= available;
			address += available;

			/*
			 * we will need another quad. Get it set up and linked
			 * into the list. It has the same command as the first.
			 */
			fq->pcommand = p->pcommand;
			fq->plink    = p->plink;
			p->plink     = fq;

			/*
			 * set the continuation bit on the previous quad
			 */
			p->pcommand |= CIO_CONT;

			/*
			 * now work on this quad...
			 */
			p = fq++;
		}
	}

	/*
	 * second pass over chain...
	 *
	 * Set byte mode for transfers if needed. If any quad in the 
	 * chain does byte transfers all quads are forced to byte mode.
	 * Check all pointers for alignment and convert them to physical 
	 * addresses. Set any semaphores that we find and flush all quads and 
	 * data buffers from the cache.
	 */

	for (p = q; !((int) p & 1); p = next_link) {
		/*
		 * check alignment of this quad
		 */
		if ((int) p & (CIO_QUAD_ALIGN - 1)) {
			printf("prepchain: quad alignment error\n");
			return EINVAL;
		}

		/*
		 * deal with link status quads...
		 */
		if (IS_CIODMA_LNKST(p->pcommand)) {
			struct linkquad *lq = (struct linkquad *) p;
			struct semaquad  *sema = lq->psema;
			struct statquad  *stat = lq->pstat;

			/*
			 * check alignment on pointers
			 */
			if ((int) sema & (CIO_QUAD_ALIGN - 1)) {
				printf("prepchain: semaphore pointer alignment error\n");
				return EINVAL;
			}

			if ((int) stat & (CIO_QUAD_ALIGN - 1)) {
				printf("prepchain: status quad pointer alignment error\n");
				return EINVAL;
			}

			/*
			 * clear the semaphore in the semaphore quad and
			 * clear the head_next pointer.
			 */
			sema->head_sema = 1;
			sema->head_next = (struct statquad *)1;
			stat->link = (struct statquad *)0;

			/*
			 * convert the virtual pointers to physical pointers
			 */

			lq->pstat = (struct statquad *) phys_addr(lq->pstat);
			if (lq->pstat == (struct statquad *) 0) {
				printf("prepchain: phys_addr error for stat\n");
				return EINVAL;
			}

			lq->psema = (struct semaquad *) phys_addr(lq->psema);
			if (lq->psema == (struct semaquad *) 0) {
				printf("prepchain: phys_addr error for sema\n");
				return EINVAL;
			}

			/*
			 * Now flush all quads from the data cache
			 */
			flush_data_cache(stat, CIO_QUAD_SIZE);
			flush_data_cache(sema, CIO_QUAD_SIZE);

		} else if (!IS_CIODMA_PAU(p->command) && 
			   !IS_CIODMA_DIS(p->command)) {

			/*
			 * if bytemode then add it to the command
			 */
			if (bytemode)
				p->pcommand |= CIO_BYTE;

			/*
			 * Check the alignment on the data buffer
			 */
			if ((int) p->paddress & (CIO_BUFFER_ALIGN - 1)) {
				printf("prepchain: data buffer align error\n");
				return EINVAL;
			}

			/*
			 * flush the buffer from the cache, at this point
			 * the address should be a virtual address, we'll
			 * convert it next...
			 */
			flush_data_cache(p->paddress, p->pcount);

			/*
			 * convert the pointer to a physical address
			 */
			if (!(p->paddress = 
			      (caddr_t) phys_addr(p->paddress))) {
				printf("prepchain: phys_addr error in link\n");
				return EINVAL;
			}
		}

		/*
		 * since we are about to clobber plink and make it a physical
		 * pointer, save it off so that we know the next quad
		 */
		next_link = p->plink;
		end_of_chain = ((u_int) p->plink & 1);

		/*
		 * convert the link pointer to a physical pointer
		 */
		if (!(p->plink = (struct dmaquad *) 
		      phys_addr((caddr_t) ((u_int) p->plink & ~1)))) {
			printf("prepchain: phys_addr error for link\n");
			return EINVAL;
		}

		if (end_of_chain)
			p->plink = (struct dmaquad *) ((u_int) p->plink | 1);

		if (cc->debug) {
			printf("address : %08x\n", p);
			printf("\tplink    : %08x\n", p->plink);
			printf("\tpcommand : %08x\t", p->pcommand);
			dump_cio_command(p->pcommand);
			if (IS_CIODMA_LNKST(p->pcommand)) {
				printf("\tpsema    : %08x\n", p->paddress);
				printf("\tpstat    : %08x\n", p->pcount);
			} else {
				printf("\tpaddress : %08x\n", p->paddress);
				printf("\tpcount   : %08x\n", p->pcount);
				if (IS_CIODMA_WC(p->pcommand) ||
				    IS_CIODMA_DLD(p->pcommand) ||
				    IS_CIODMA_WD(p->pcommand))
					dump_buffer(p->address, p->count);
			}
		}

		/*
		 * now flush the quad itself from the data cache
		 */
		flush_data_cache(p, CIO_QUAD_SIZE);

	}
	return 0;
}

/*
 * equivalently mapped data structures for CIO drivers
 */
char	ciodatums[CIOBYTES];
char	*eciodatums = &ciodatums[CIOBYTES];

/*
 * get equivalently mapped memory with a specified alignment
 */
caddr_t
cio_malloc(size, align)
	int size;
{
	static char *bp = ciodatums;
	register char *rbp;

	rbp = (char *) ((int) bp + align - 1 & ~(align - 1));

	bp = rbp + size;
	if (bp > eciodatums)
		panic("cio_malloc: not enough memory");
	return (caddr_t) rbp;
}

dump_quads(q)
	struct dmaquad *q;
{
	char *b;
	int i;

	printf("dump_quads\n");
	while(!((int) q & 1)) {
		printf("address : %08x\n", q);
		printf("\tlink    : %08x\n", q->link);
		printf("\tcommand : %08x\t", q->command);
		dump_cio_command(q->command);
		if (IS_CIODMA_LNKST(q->command)) {
			printf("\tsema    : %08x\n", q->address);
			printf("\tstat    : %08x\n", q->count);
		} else {
			printf("\taddress : %08x\n", q->address);
			printf("\tcount   : %08x\n", q->count);
			if (IS_CIODMA_WC(q->command) ||
			    IS_CIODMA_DLD(q->command) ||
			    IS_CIODMA_WD(q->command))
				dump_buffer(q->address, q->count);
		}
		q = q->link;
	}
	printf("\n");
}

dump_cc(cc)
	struct cio_chan *cc;
{
	printf("cc : %x\n", cc);
	printf("\ttype    : %x", cc->type);

	switch (cc->type) {
	case CIOIDY_HPIB:
		printf(" hpib\n");
		break;
	case CIOIDY_LAN:
		printf(" lan\n");
		break;
	case CIOIDY_MUX:
		printf(" mux\n");
		break;
	default:
		printf(" unknown\n");
		break;
	}

	printf("\tciounit  = %x\n", cc->ciounit);
	printf("\tchannel  = %x\n", cc->channel);
	printf("\talive    = %x\n", cc->alive);
	printf("\tnotified = %x\n", cc->notified);
	printf("\tint_level= %x\n", cc->int_level);
	printf("\tdevice   = %x\n", cc->device);
	printf("\tdio      = %x\n", cc->dio);
	printf("\tsubc     = %x\n", cc->subc);
	printf("\tchain    = %x\n", cc->chain);
	printf("\thpa      = %x\n", cc->hpa);
	printf("\tactive   = %x\n", cc->active);
	printf("\tdebug    = %x\n\n", cc->debug);
}

dump_cr(cr)
	struct cio_request *cr;
{

	printf("cr = %x\n", cr);
	printf("cr->cmdq    = %x\n", cr->cmdq);
	printf("cr->stat    = %x\n", cr->stat);
	printf("cr->quad    = %x\n", cr->quad);
	printf("cr->ciounit = %x\n", cr->ciounit);
	printf("cr->channel = %x\n", cr->channel);
}

dump_cio_command(c)
{

	switch(c >> 28 & 0x0f) {
	case 0x0:
		if (IS_CIODMA_IDY(c))
			printf(" CIODMA_IDY");
		else if (IS_CIODMA_LNKST(c))
			printf(" CIODMA_LNKST");
		else {
			printf("\n");
			return;
		}
		break;

	case 0x1:
		printf(" CIODMA_DLD");
		if (c & CIO_CODE)
			printf(" | CIO_CODE");
		break;

	case 0x2:
		printf(" CIODMA_PAU");
		break;

	case 0x3:
		printf(" CIODMA_DIS");
		break;

	case 0x4:
		printf(" CIODMA_RS");
		if (c & CIO_DISC)
			printf(" | CIO_DISC");
		break;

	case 0x5:
		printf(" CIODMA_WC");
		break;

	case 0x6:
		printf(" CIODMA_RD");
		break;

	case 0x7:
		printf(" CIODMA_WD");
		break;

	case 0x8:
		printf(" CIODMA_RTS");
		break;

	case 0x9:
		printf(" CIODMA_WTC");
		break;

	case 0xA:
		printf(" CIODMA_RDS");
		if (c & CIO_CODE)
			printf(" | CIO_CODE");
		break;

	case 0xB:
		printf(" CIODMA_CLC");
		break;

	default:
		printf("\n");
		return;
	}


	if (c & CIO_BYTE)
		printf(" | CIO_BYTE");

	if (c & CIO_INTR)
		printf(" | CIO_INTR");

	if (c & CIO_CONT)
		printf(" | CIO_CONT");

	printf("\n");
}

dump_buffer(b, c) 
	char *b;
{
	int i;
	printf("dump_buffer: b = %x\n", b);

	if (c > 64)
		c = 64;

	for (i = 0; i < c; i++) {
		if (i % 16 == 0)
			printf("\t");
		printf("%02x%c", *b++ & 0xff, (i % 16 == 15) ? '\n' : ' ');
	}
	if (c && ((c - 1) % 16 != 15))
		printf("\n");
}

dump_status(cc)
	struct cio_chan *cc;
{
	int i;

	printf("         cc->hpa->io_status = %x\n", cc->hpa->io_status);
	printf("         cc->hpa->io_status = %x\n", cc->hpa->io_status);
	printf("         cc->hpa->interrupt = %x\n", cc->hpa->interrupt);
	printf("         cc->hpa->sub_mask_set = %x\n", cc->hpa->sub_mask_set);
	printf("         cc->subc->currquad_link = %x\n",
	       cc->subc->currquad_link);
	printf("         cc->subc->currquad_cmd = %x\n",
	       cc->subc->currquad_cmd);
	printf("         cc->subc->currquad_addr = %x\n",
	       cc->subc->currquad_addr);
	printf("         cc->subc->currquad_count = %x\n",
	       cc->subc->currquad_count);
	printf("         cc->subc->status = %x\n", cc->subc->status);
}


#endif NCIO > 0

